from unittest import mock
import uuid

from jsonschema import SchemaError
import pytest

from hedwig.exceptions import ValidationError
from hedwig.models import MessageType
from hedwig.validator import MessageValidator
from hedwig.testing.factories import MessageFactory


class TestMessageValidator:
    def setup_method(self, method):
        self.validator = MessageValidator()

    @mock.patch('hedwig.validator.MessageValidator.check_schema')
    def test_constructor_checks_schema(self, mock_check_schema):
        schema = {'schemas': []}
        MessageValidator(schema)
        mock_check_schema.assert_called_once_with(schema)

    @pytest.mark.parametrize(
        'schema,schema_exc_error',
        [
            [
                {
                    'schemas': []
                },
                "Invalid schema file: expected key 'schemas' with non-empty value",
            ],
            [
                {
                    'schemas': {
                        "device.created": []
                    }
                },
                "Invalid definition for message type: 'device.created', value must contain a dict of valid versions",
            ],
            [
                {
                    'schemas': {
                        "device.created": {
                            'fail-pattern': []
                        }
                    }
                },
                "Invalid version 'fail-pattern' for message type: 'device.created'",
            ],
            [
                {
                    'schemas': {
                        "device.created": {
                            '1.*': []
                        }
                    }
                },
                "Invalid version '1.*' for message type: 'device.created'",
            ],
            [
                {
                    'schemas': {
                        "device.created": {
                            '1.*': {},
                        },
                        "vehicle_created": {
                            '1.*': {},
                        },
                        "trip_created": {
                            '1.*': {},
                        }
                    }
                },
                "Schema not found for 'trip_created' v2.*",
            ],
        ]
    )
    def test_check_schema(self, schema, schema_exc_error):
        with pytest.raises(SchemaError) as exc_context:
            MessageValidator(schema)

        assert schema_exc_error in exc_context.value.args[0]

    def test_validate(self):
        message = MessageFactory(msg_type=MessageType.trip_created, model_version=1)
        self.validator.validate(message)

    def test_validate_raises_error_invalid_schema(self):
        with pytest.raises(ValidationError):
            MessageFactory(schema='https://wrong.host/schema#/schemas/trip_created/1.0')

        with pytest.raises(ValidationError):
            MessageFactory(schema='https://hedwig.automatic.com/schema#/schemas/trip_created/9.0')

    def test_validate_raises_errors(self):
        with pytest.raises(ValidationError):
            MessageFactory(data={})

    def test_check_human_uuid(self):
        assert self.validator.check_human_uuid(str(uuid.uuid4()))
        assert self.validator.check_human_uuid('6cac5588-24cc-4b4f-bbf9-7dc0ce93f96e')
        assert self.validator.check_human_uuid('793befdf-56c2-416a-92ed-420e62b33eb5')

    def test_check_human_uuid_fails(self):
        assert not self.validator.check_human_uuid(uuid.uuid4().hex)
        assert not self.validator.check_human_uuid('6cac5588')
        assert not self.validator.check_human_uuid('yyyyyyyy-tttt-416a-92ed-420e62b33eb5')
        assert not self.validator.check_human_uuid(uuid.uuid4())


def test_custom_validator(settings):
    settings.HEDWIG_DATA_VALIDATOR_CLASS = 'tests.validator.CustomValidator'

    with pytest.raises(ValidationError):
        MessageFactory(msg_type=MessageType.trip_created, addition_version=1, data__vin='o' * 17)



import pprint
from distutils.version import StrictVersion
from unittest import mock

import pytest


__all__ = ['mock_hedwig_publish']


@pytest.fixture()
def mock_hedwig_publish():
    from hedwig.publisher import publish

    with mock.patch('hedwig.publisher.publish', wraps=publish) as mock_publish, mock.patch(
        'hedwig.publisher._publish_over_sns'
    ):

        def assert_message_published(msg_type, data=None, version=StrictVersion('1.0')):
            """
            Helper function to check if a Hedwig message with given type, data
            and schema version was sent.
            """
            if not isinstance(version, StrictVersion):
                version = StrictVersion(version)

            assert any(
                msg.type == msg_type and (data is None or msg.data) == data and msg.data_schema_version == version
                for (msg,), _ in mock_publish.call_args_list
            ), pprint.pformat(
                [(msg.type, msg.data, msg.data_schema_version) for (msg,), _ in mock_publish.call_args_list]
            )

        def assert_message_not_published(msg_type):
            """
            Helper function to check that a Hedwig message of given type was NOT sent.
            """
            assert all(msg.type != msg_type for (msg,), _ in mock_publish.call_args_list), pprint.pformat(
                [(msg.type, msg.data, msg.data_schema_version) for (msg,), _ in mock_publish.call_args_list]
            )

        mock_publish.assert_message_published = assert_message_published
        mock_publish.assert_message_not_published = assert_message_not_published
        yield mock_publish



import pytest

import hedwig.conf


def test_fail_import_bad_value(settings):
    settings.HEDWIG_DEFAULT_HEADERS = 'foo'

    with pytest.raises(ImportError):
        hedwig.conf.settings.HEDWIG_DEFAULT_HEADERS


def test_fail_import_bad_module(settings):
    settings.HEDWIG_DEFAULT_HEADERS = 'foo.bar'

    with pytest.raises(ImportError):
        hedwig.conf.settings.HEDWIG_DEFAULT_HEADERS


def test_fail_import_bad_attr(settings):
    settings.HEDWIG_DEFAULT_HEADERS = 'tests.validator.foobar'

    with pytest.raises(ImportError):
        hedwig.conf.settings.HEDWIG_DEFAULT_HEADERS



import json
import itertools
import logging
import threading
import typing

import boto3

from hedwig.conf import settings
from hedwig.exceptions import RetryException, IgnoreException, ValidationError, LoggingException
from hedwig.models import Message


WAIT_TIME_SECONDS = 20  # Maximum allowed by SQS

logger = logging.getLogger(__name__)


def _get_sqs_resource():
    return boto3.resource(
        'sqs',
        region_name=settings.AWS_REGION,
        aws_access_key_id=settings.AWS_ACCESS_KEY,
        aws_secret_access_key=settings.AWS_SECRET_KEY,
        aws_session_token=settings.AWS_SESSION_TOKEN,
        endpoint_url=settings.AWS_ENDPOINT_SQS,
    )


def get_default_queue_name() -> str:
    return f'HEDWIG-{settings.HEDWIG_QUEUE}'


def log_received_message(message_body: dict) -> None:
    logger.debug('Received message', extra={
        'message_body': message_body,
    })


def log_invalid_message(message_json: str) -> None:
    logger.debug('Received invalid message', extra={
        'message_json': message_json,
    })


def _load_and_validate_message(data: dict) -> Message:
    message = Message(data)
    message.validate_callback()

    return message


def message_handler(message_json: str, receipt: typing.Optional[str]) -> None:
    try:
        message_body = json.loads(message_json)
        settings.HEDWIG_POST_DESERIALIZE_HOOK(message_data=message_body)
        message = _load_and_validate_message(message_body)
    except (ValidationError, ValueError):
        log_invalid_message(message_json)
        raise

    log_received_message(message_body)

    if receipt is not None:
        message.metadata.receipt = receipt

    try:
        message.exec_callback()
    except IgnoreException:
        logger.info(f'Ignoring task {message.id}')
        return
    except LoggingException as e:
        # log with message and extra
        logger.exception(str(e), extra=e.extra)
        # let it bubble up so message ends up in DLQ
        raise
    except RetryException:
        # Retry without logging exception
        logger.info('Retrying due to exception')
        # let it bubble up so message ends up in DLQ
        raise
    except Exception:
        logger.exception(f'Exception while processing message')
        # let it bubble up so message ends up in DLQ
        raise


def message_handler_sqs(queue_message) -> None:
    message_json = queue_message.body
    receipt = queue_message.receipt_handle

    message_handler(message_json, receipt)


def message_handler_lambda(lambda_record: dict) -> None:
    message_json = lambda_record['Sns']['Message']
    receipt = None

    message_handler(message_json, receipt)


def get_queue(queue_name: str):
    sqs = _get_sqs_resource()
    return sqs.get_queue_by_name(QueueName=queue_name)


def get_queue_messages(queue, num_messages: int, wait_timeout_s: typing.Optional[int] = None,
                       visibility_timeout: typing.Optional[int] = None) -> list:
    params = {
        'MaxNumberOfMessages': num_messages,
        'WaitTimeSeconds': wait_timeout_s or WAIT_TIME_SECONDS,
        'MessageAttributeNames': ['All'],
    }
    if visibility_timeout is not None:
        params['VisibilityTimeout'] = visibility_timeout
    return queue.receive_messages(**params)


def fetch_and_process_messages(
        queue_name: str, queue, num_messages: int = 1, visibility_timeout: typing.Optional[int] = None) -> None:

    for queue_message in get_queue_messages(queue, num_messages=num_messages, visibility_timeout=visibility_timeout):
        settings.HEDWIG_PRE_PROCESS_HOOK(sqs_queue_message=queue_message)

        try:
            message_handler_sqs(queue_message)
            try:
                queue_message.delete()
            except Exception:
                logger.exception(f'Exception while deleting message from {queue_name}')
        except Exception:
            # already logged in message_handler
            pass


def process_messages_for_lambda_consumer(lambda_event: dict) -> None:
    for record in lambda_event['Records']:
        settings.HEDWIG_PRE_PROCESS_HOOK(sns_record=record)

        message_handler_lambda(record)


def listen_for_messages(
        num_messages: int = 10, visibility_timeout_s: typing.Optional[int] = None,
        loop_count: typing.Optional[int] = None, shutdown_event: threading.Event = None) -> None:
    """
    Starts a Hedwig listener for message types provided and calls the callback handlers like so:

    callback_fn(message).

    The message is explicitly deleted only if callback function ran successfully. In case of an exception the message is
    kept on queue and processed again. If the callback keeps failing, SQS dead letter queue mechanism kicks in and
    the message is moved to the dead-letter queue.

    This function is blocking by default. It may be run for specific number of loops by passing `loop_count`. It may
    also be stopped by passing a shut down event object which can be set to stop the function.

    :param num_messages: Maximum number of messages to fetch in one SQS API call. Defaults to 10
    :param visibility_timeout_s: The number of seconds the message should remain invisible to other queue readers.
        Defaults to None, which is queue default
    :param loop_count: How many times to fetch messages from SQS. Default to None, which means loop forever.
    :param shutdown_event: An event to signal that the process should shut down. This prevents more messages from
        being de-queued and function exits after the current messages have been processed.
    """
    if not shutdown_event:
        shutdown_event = threading.Event()

    queue_name = get_default_queue_name()

    queue = get_queue(queue_name)
    for count in itertools.count():
        if (loop_count is None or count < loop_count) and not shutdown_event.is_set():
            fetch_and_process_messages(
                queue_name, queue, num_messages=num_messages, visibility_timeout=visibility_timeout_s)
        else:
            break



import json
import re
import typing
from distutils.version import StrictVersion
from pathlib import Path

import funcy
from jsonschema import SchemaError, RefResolutionError, FormatChecker
from jsonschema.validators import Draft4Validator

from hedwig.conf import settings
from hedwig.exceptions import ValidationError

if typing.TYPE_CHECKING:
    from hedwig.models import Message  # noqa  # pragma: no cover


class MessageValidator(Draft4Validator):
    # uuid separated by hyphens:
    _human_uuid_re = re.compile("^[0-9a-f]{8}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{12}$")

    _version_pattern_re = re.compile("^[0-9]+\.\*$")

    checker = FormatChecker()
    """
    FormatChecker that checks for `format` JSON-schema field. This may be customized by an app by overriding setting
    `HEDWIG_DATA_VALIDATOR_CLASS` and defining more format checkers.
    """

    def __init__(self, schema: typing.Optional[dict] = None) -> None:
        if schema is None:
            # automatically load schema
            schema_filepath = settings.HEDWIG_SCHEMA_FILE
            with open(schema_filepath) as f:
                schema = json.load(f)

        self.check_schema(schema)

        super(MessageValidator, self).__init__(schema, format_checker=self.checker)

    @property
    def schema_root(self) -> str:
        return self.schema['id']

    def validate(self, message: 'Message') -> None:
        """
        Validates a message using JSON Schema
        """
        if not message.schema.startswith(self.schema_root):
            raise ValidationError

        full_version = StrictVersion(message.schema.split('/')[-1])
        major_version = full_version.version[0]
        schema_ptr = message.schema.replace(str(full_version), str(major_version)) + '.*'

        try:
            _, schema = self.resolver.resolve(schema_ptr)
        except RefResolutionError:
            raise ValidationError

        errors = list(self.iter_errors(message.data, schema))
        if errors:
            raise ValidationError(errors)

    @classmethod
    def check_schema(cls, schema: dict) -> None:
        msg_types_found = {k: False for k in funcy.chain(settings.HEDWIG_MESSAGE_ROUTING, settings.HEDWIG_CALLBACKS)}
        # custom validation for Hedwig - TODO ideally should just be represented in json-schema file as meta schema,
        # however jsonschema lib doesn't support draft 06 which is what's really required here
        errors = []
        if not schema.get('schemas'):
            errors.append("Invalid schema file: expected key 'schemas' with non-empty value")
        else:
            for msg_type, versions in schema['schemas'].items():
                if not isinstance(versions, dict) or not versions:
                    errors.append(f"Invalid definition for message type: '{msg_type}', value must contain a dict of "
                                  f"valid versions")
                else:
                    for version_pattern, definition in versions.items():
                        if not cls._version_pattern_re.match(version_pattern):
                            errors.append(f"Invalid version '{version_pattern}' for message type: '{msg_type}'")
                        if (msg_type, version_pattern) in msg_types_found:
                            msg_types_found[(msg_type, version_pattern)] = True
                        if not isinstance(definition, dict) and not definition:
                            errors.append(f"Invalid version '{version_pattern}' for message type: '{msg_type}'")

        for (msg_type, version_pattern), found in msg_types_found.items():
            if not found:
                errors.append(f"Schema not found for '{msg_type}' v{version_pattern}")

        if errors:
            raise SchemaError(errors)

    @staticmethod
    @FormatChecker.cls_checks('human-uuid')
    def check_human_uuid(instance):
        if not isinstance(instance, str):
            return False
        return bool(MessageValidator._human_uuid_re.match(instance))


class FormatValidator(Draft4Validator):
    def __init__(self):
        # automatically load schema
        schema_filepath = Path(__file__).resolve().parent / 'format_schema.json'
        with open(schema_filepath) as f:
            schema = json.load(f)

        super(FormatValidator, self).__init__(schema)

    def validate(self, data: dict) -> None:
        errors = list(self.iter_errors(data))
        if errors:
            raise ValidationError(errors)



import copy
import re
import time
import typing
import uuid
from distutils.version import StrictVersion
from enum import Enum

import boto3

from hedwig.conf import settings
from hedwig.exceptions import ValidationError, CallbackNotFound
from hedwig.validator import FormatValidator


MessageType = Enum(    # type: ignore
    'MessageType',
    {
        t[0].replace('.', '_').replace('-', '_'): t[0]
        for t in settings.HEDWIG_MESSAGE_ROUTING
    }
)

MessageType.__doc__ = "Enumeration representing the message types supported for this service. This is automatically " \
                      "created based on setting `HEDWIG_MESSAGE_ROUTING`"

_validator = None


def _get_validator():
    global _validator
    if _validator is None:
        _validator = settings.HEDWIG_DATA_VALIDATOR_CLASS()
    return _validator


_format_validator = FormatValidator()


class Metadata:
    def __init__(self, data: dict) -> None:
        self._timestamp = data['timestamp']
        self._publisher = data['publisher']
        self._headers = data['headers']
        self._receipt: typing.Optional[str] = None

    @property
    def timestamp(self) -> int:
        """
        Timestamp of message creation in epoch milliseconds
        """
        return self._timestamp

    @property
    def publisher(self) -> str:
        """
        Publisher of message
        """
        return self._publisher

    @property
    def receipt(self) -> typing.Optional[str]:
        """
        SQS receipt for the task. This may be used to extend message visibility if the task is running longer
        than expected using :meth:`Message.extend_visibility_timeout`
        """
        return self._receipt

    @receipt.setter
    def receipt(self, value: str) -> None:
        """
        SQS receipt of the message
        """
        assert isinstance(value, str)
        self._receipt = value

    @property
    def headers(self) -> dict:
        """
        Custom headers sent with the message
        """
        return self._headers

    def as_dict(self) -> dict:
        return {
            "timestamp": self.timestamp,
            "headers": self.headers,
            "publisher": self.publisher,
        }

    def __eq__(self, other) -> bool:
        if not isinstance(other, self.__class__):
            return NotImplemented
        return self.as_dict() == typing.cast(Metadata, other).as_dict()


def _get_sqs_client():
    return boto3.client(
        'sqs',
        region_name=settings.AWS_REGION,
        aws_access_key_id=settings.AWS_ACCESS_KEY,
        aws_secret_access_key=settings.AWS_SECRET_KEY,
        aws_session_token=settings.AWS_SESSION_TOKEN,
        endpoint_url=settings.AWS_ENDPOINT_SQS,
    )


def _get_queue_url(client, queue_name: str) -> str:
    response = client.get_queue_url(
        QueueName=queue_name,
    )
    return response['QueueUrl']


class Message:
    """
    Model for Hedwig messages. All properties of a message should be considered immutable.
    A Message object will always have known message format schema and message format schema version even if the data
    _may_ not be valid.
    """
    FORMAT_CURRENT_VERSION = StrictVersion('1.0')
    FORMAT_VERSIONS = [StrictVersion('1.0')]
    '''
    Here are the schema definitions:

    Version 1.0:
    {
        "format_version": "1.0",
        "schema": "https://hedwig.automatic.com/schema#/schemas/trip.created/1.0",
        "id": "b1328174-a21c-43d3-b303-964dfcc76efc",
        "metadata": {
            "timestamp": 1460868253255,
            "publisher": "myapp",
            "headers": {
                ...
            }
        },
        "data": {
            ...
        }
    }

    All the top-level fields (other than `metadata`) are required to be non-empty. `metadata` field is expected to
    be present, but may be empty. All fields in `metadata` are optional. `data` is validated using `schema`.
    '''

    # schema parsing re, eg: hedwig.automatic.com/schema#/schemas/trip.created/1.0
    _schema_re = re.compile(r'([^/]+)/([^/]+)$')

    def __init__(self, data: dict) -> None:
        """
        See message format version definitions above
        """
        _format_validator.validate(data)

        self._id = data['id']
        self._metadata = Metadata(data['metadata'])
        self._schema = data['schema']
        self._data = data['data']
        self._format_version = StrictVersion(data['format_version'])

        self.validate()

    def validate(self) -> None:
        """
        Validates a message using JSON schema.

        :raise: :class:`hedwig.ValidationError` if validation fails.
        """
        try:
            mo = self._schema_re.search(self.schema)
            if mo is None:
                raise ValueError
            schema_groups = mo.groups()
            self._type = MessageType(schema_groups[0])
            self._data_schema_version = StrictVersion(schema_groups[1])
        except (AttributeError, ValueError):
            raise ValidationError(f'Invalid schema found: {self.schema}')

        _get_validator().validate(self)

    def validate_callback(self) -> None:
        from hedwig.callback import Callback

        try:
            self._callback = Callback.find_by_message(self.type, self.data_schema_version.version[0])
        except CallbackNotFound:
            raise ValidationError

    def exec_callback(self) -> None:
        """
        Call the callback with this message
        """
        self.callback.call(self)

    @classmethod
    def _create_metadata(cls, headers: dict) -> dict:
        return {
            'timestamp': int(time.time() * 1000),
            'publisher': settings.HEDWIG_PUBLISHER,
            'headers': headers,
        }

    @classmethod
    def new(cls, msg_type: MessageType, data_schema_version: StrictVersion, data: dict, msg_id: str=None,
            headers: dict=None) -> 'Message':
        """
        Creates Message object given type, data schema version and data. This is typically used by the publisher code.

        :param msg_type: MessageType instance
        :param data_schema_version: StrictVersion representing data schema
        :param data: The dict to pass in `data` field of Message.
        :param msg_id: Custom message identifier. If not passed, a randomly generated uuid will be used.
        :param headers: Custom headers
        """
        assert isinstance(msg_type, MessageType)
        assert isinstance(data_schema_version, StrictVersion)
        assert isinstance(data, dict)
        assert isinstance(msg_id, (type(None), str))
        assert isinstance(headers, (type(None), dict))

        return Message(data={
            'format_version': str(cls.FORMAT_CURRENT_VERSION),
            'id': msg_id or str(uuid.uuid4()),
            'schema': f'{_get_validator().schema_root}#/schemas/{msg_type.value}/{data_schema_version}',
            'metadata': cls._create_metadata(headers or {}),
            'data': copy.deepcopy(data),
        })

    def publish(self):
        """
        Publish this message on Hedwig infra
        """
        from hedwig.publisher import publish

        publish(self)

    def extend_visibility_timeout(self, visibility_timeout_s: int) -> None:
        """
        Extends visibility timeout of a message for long running tasks.
        """
        from hedwig.consumer import get_default_queue_name

        queue_name = get_default_queue_name()

        client = _get_sqs_client()

        queue_url = _get_queue_url(client, queue_name)

        client.change_message_visibility(
            QueueUrl=queue_url,
            ReceiptHandle=self.receipt,
            VisibilityTimeout=visibility_timeout_s,
        )

    def __eq__(self, other) -> bool:
        if not isinstance(other, self.__class__):
            return NotImplemented
        return self.as_dict() == typing.cast(Message, other).as_dict()

    @property
    def data_schema_version(self) -> StrictVersion:
        """
        `StrictVersion` object representing data schema version. May be `None` if message can't be validated.
        """
        return self._data_schema_version

    @property
    def callback(self):
        return self._callback

    @property
    def id(self) -> str:
        """
        Message identifier
        """
        return self._id

    @property
    def schema(self) -> str:
        """
        Message schema
        """
        return self._schema

    @property
    def type(self) -> MessageType:
        """
        MessageType. May be none if message is invalid
        """
        return self._type

    @property
    def format_version(self) -> StrictVersion:
        """
        Message format version (this is different from data schema version)
        """
        return self._format_version

    @property
    def metadata(self) -> Metadata:
        """
        Message metadata
        """
        return self._metadata

    @property
    def timestamp(self) -> int:
        return self.metadata.timestamp

    timestamp.__doc__ = Metadata.timestamp.__doc__

    @property
    def headers(self) -> dict:
        return self._metadata.headers

    headers.__doc__ = Metadata.headers.__doc__

    @property
    def receipt(self) -> typing.Optional[str]:
        return self.metadata.receipt

    receipt.__doc__ = Metadata.receipt.__doc__

    @property
    def publisher(self) -> typing.Optional[str]:
        return self.metadata.publisher

    publisher.__doc__ = Metadata.publisher.__doc__

    @property
    def data(self) -> dict:
        """
        Message data
        """
        return self._data

    @property
    def topic(self) -> str:
        """
        The SNS topic name for routing the message
        """
        version_pattern = f'{self.data_schema_version.version[0]}.*'
        return settings.HEDWIG_MESSAGE_ROUTING[(self.type.value, version_pattern)]

    def as_dict(self) -> dict:
        return {
            'id': self.id,
            'format_version': str(self.format_version),
            'schema': self.schema,
            'metadata': self.metadata.as_dict(),
            'data': self.data,
        }



import re

from hedwig.validator import MessageValidator


class CustomValidator(MessageValidator):
    # simplistic check: 17 alphanumeric characters except i, o, q
    _vin_re = re.compile("^[a-hj-npr-z0-9]{17}$")

    @staticmethod
    @MessageValidator.checker.checks('vin')
    def check_vin(instance) -> bool:
        if not isinstance(instance, str):
            return True
        return bool(CustomValidator._vin_re.match(instance))



from decimal import Decimal
import json
from unittest import mock
import uuid

import pytest

from hedwig.conf import settings
from hedwig.models import MessageType
from hedwig.exceptions import ValidationError, CallbackNotFound
from hedwig.publisher import publish, _get_sns_topic, _convert_to_json, _publish_over_sns, _get_sns_client
from hedwig.testing.factories import MessageFactory


@mock.patch('hedwig.publisher.boto3.client', autospec=True)
def test_get_sns_client(mock_boto3_client):
    client = _get_sns_client()
    mock_boto3_client.assert_called_once_with(
        'sns',
        region_name=settings.AWS_REGION,
        aws_access_key_id=settings.AWS_ACCESS_KEY,
        aws_secret_access_key=settings.AWS_SECRET_KEY,
        aws_session_token=settings.AWS_SESSION_TOKEN,
        endpoint_url=settings.AWS_ENDPOINT_SNS,
        config=mock.ANY,
    )
    assert client == mock_boto3_client.return_value


def test__get_sns_topic(message):
    message.validate()
    assert _get_sns_topic(message) == f'arn:aws:sns:{settings.AWS_REGION}:{settings.AWS_ACCOUNT_ID}:hedwig-' \
                                      f'{message.topic}'


@mock.patch('hedwig.publisher._get_sns_client', autospec=True)
def test__publish_over_sns(mock_get_sns_client, message):
    message.validate()
    topic = _get_sns_topic(message)
    message_json = _convert_to_json(message.as_dict())

    _publish_over_sns(topic, message_json, message.headers)

    mock_get_sns_client.assert_called_once_with()
    mock_get_sns_client.return_value.publish.assert_called_once_with(
        TopicArn=topic,
        Message=message_json,
        MessageAttributes={
            k: {
                'DataType': 'String',
                'StringValue': str(v),
            } for k, v in message.headers.items()
        },
    )


@pytest.mark.parametrize('value', [1469056316326, 1469056316326.123])
def test__convert_to_json_decimal(value, message_data):
    message_data['data']['vin'] = Decimal(value)
    assert json.loads(_convert_to_json(message_data))['data']['vin'] == float(message_data['data']['vin'])


def test__convert_to_json_non_serializable(message_data):
    message_data['data']['vin'] = object()
    with pytest.raises(TypeError):
        _convert_to_json(message_data)


@mock.patch('hedwig.publisher._convert_to_json', autospec=True)
@mock.patch('hedwig.publisher._publish_over_sns', autospec=True)
def test_publish(mock_publish_over_sns, mock_convert_to_json, message):
    message.validate()

    sns_id = str(uuid.uuid4())

    mock_publish_over_sns.return_value = {'MessageId': sns_id}

    publish(message)

    topic = _get_sns_topic(message)
    mock_publish_over_sns.assert_called_once_with(
        topic,
        mock_convert_to_json.return_value,
        message.headers
    )
    mock_convert_to_json.assert_called_once_with(message.as_dict())


default_headers_hook = mock.MagicMock()


@mock.patch('hedwig.publisher._convert_to_json', autospec=True)
@mock.patch('hedwig.publisher._publish_over_sns', autospec=True)
def test_default_headers_hook(mock_publish_over_sns, mock_convert_to_json, message, settings):
    settings.HEDWIG_DEFAULT_HEADERS = 'tests.test_publisher.default_headers_hook'
    default_headers_hook.return_value = {
        'mickey': 'mouse',
    }

    message.validate()

    sns_id = str(uuid.uuid4())

    mock_publish_over_sns.return_value = {'MessageId': sns_id}

    publish(message)

    topic = _get_sns_topic(message)
    mock_publish_over_sns.assert_called_once_with(
        topic,
        mock_convert_to_json.return_value,
        {**message.headers, **default_headers_hook.return_value},
    )
    expected = message.as_dict()
    expected['metadata']['headers'].update(default_headers_hook.return_value)
    mock_convert_to_json.assert_called_once_with(expected)

    default_headers_hook.assert_called_once_with(message=message)


def pre_serialize_hook(message_data):
    # clear headers to make sure we are not able to destroy sqs
    # message attributes
    message_data['metadata']['headers'].clear()


@mock.patch('hedwig.publisher._convert_to_json', autospec=True)
@mock.patch('hedwig.publisher._publish_over_sns', autospec=True)
def test_pre_serialize_hook(mock_publish_over_sns, mock_convert_to_json, message, settings):
    settings.HEDWIG_PRE_SERIALIZE_HOOK = 'tests.test_publisher.pre_serialize_hook'

    message.validate()

    sns_id = str(uuid.uuid4())

    mock_publish_over_sns.return_value = {'MessageId': sns_id}

    publish(message)

    topic = _get_sns_topic(message)
    mock_publish_over_sns.assert_called_once_with(
        topic,
        mock_convert_to_json.return_value,
        message.headers
    )
    message_data = message.as_dict()
    message_data['metadata']['headers'].clear()
    mock_convert_to_json.assert_called_once_with(message_data)


@mock.patch('tests.handlers._trip_created_handler', autospec=True)
def test_sync_mode(callback_mock, message, settings):
    settings.HEDWIG_SYNC = True

    message.publish()
    callback_mock.assert_called_once_with(message)


def test_sync_mode_detects_invalid_callback(settings):
    settings.HEDWIG_SYNC = True

    message = MessageFactory(msg_type=MessageType.vehicle_created)
    with pytest.raises(ValidationError) as exc_info:
        message.publish()
    assert isinstance(exc_info.value.__context__, CallbackNotFound)



from unittest import mock
import uuid

import pytest

from hedwig.callback import Callback
from hedwig.exceptions import ConfigurationError, CallbackNotFound
from hedwig.models import Message, MessageType
from tests.handlers import trip_created_handler


def default_headers() -> dict:
    return {
        'request_id': str(uuid.uuid4()),
    }


class TestCallback:
    @staticmethod
    def f(message):
        pass

    @staticmethod
    def f_kwargs(**kwargs):
        pass

    @staticmethod
    def f_args(message, *args, **kwargs):
        pass

    @staticmethod
    def f_invalid_annotation(message: dict):
        pass

    @staticmethod
    def f_no_param():
        pass

    @staticmethod
    def f_unknown_param(message, unknown):
        pass

    def test_constructor(self):
        task_obj = Callback(TestCallback.f)
        assert task_obj.fn is TestCallback.f

    def test_constructor_disallow_kwargs(self):
        with pytest.raises(ConfigurationError):
            Callback(TestCallback.f_kwargs)

    def test_constructor_disallow_args(self):
        with pytest.raises(ConfigurationError):
            Callback(TestCallback.f_args)

    def test_constructor_bad_annotation(self):
        with pytest.raises(ConfigurationError):
            Callback(TestCallback.f_invalid_annotation)

    def test_constructor_no_param(self):
        with pytest.raises(ConfigurationError):
            Callback(TestCallback.f_no_param)

    def test_constructor_unknown_param(self):
        with pytest.raises(ConfigurationError):
            Callback(TestCallback.f_unknown_param)

    def test_call(self, message):
        _f = mock.MagicMock()

        def f(message: Message):
            _f(message)

        Callback(f).call(message)
        _f.assert_called_once_with(message)

    def test_find_by_message(self):
        assert Callback.find_by_message(MessageType.trip_created, 1)._fn == trip_created_handler

    def test_find_by_name_fail(self):
        with pytest.raises(CallbackNotFound):
            Callback.find_by_message(MessageType.vehicle_created, 1)

    def test_str(self):
        assert str(Callback(self.f)) == 'Hedwig task: f'



import json
from unittest import mock
import uuid

import pytest

from hedwig import consumer
from hedwig.consumer import (
    get_queue, _load_and_validate_message, get_default_queue_name, message_handler, listen_for_messages,
    message_handler_lambda, process_messages_for_lambda_consumer, message_handler_sqs, get_queue_messages,
    WAIT_TIME_SECONDS, fetch_and_process_messages, _get_sqs_resource,
)
from hedwig.conf import settings
from hedwig.exceptions import RetryException, IgnoreException, ValidationError, LoggingException


@mock.patch('hedwig.consumer.boto3.resource', autospec=True)
def test__get_sqs_resource(mock_boto3_resource):
    resource = _get_sqs_resource()
    mock_boto3_resource.assert_called_once_with(
        'sqs',
        region_name=settings.AWS_REGION,
        aws_access_key_id=settings.AWS_ACCESS_KEY,
        aws_secret_access_key=settings.AWS_SECRET_KEY,
        aws_session_token=settings.AWS_SESSION_TOKEN,
        endpoint_url=settings.AWS_ENDPOINT_SQS,
    )
    assert resource == mock_boto3_resource.return_value


@mock.patch('hedwig.consumer._get_sqs_resource', autospec=True)
def test_get_queue(mock_get_sqs_resource):
    queue_name = 'foo'
    assert mock_get_sqs_resource.return_value.get_queue_by_name.return_value == get_queue(queue_name)
    mock_get_sqs_resource.assert_called_once_with()
    mock_get_sqs_resource.return_value.get_queue_by_name.assert_called_once_with(QueueName=queue_name)


@mock.patch('hedwig.consumer.Message', autospec=True)
def test__load_and_validate_message(mock_message, message_data):
    _load_and_validate_message(message_data)
    mock_message.assert_called_once_with(message_data)


@mock.patch('hedwig.consumer.Message.exec_callback', autospec=True)
@mock.patch('hedwig.consumer._load_and_validate_message', autospec=True)
class TestMessageHandler:
    def test_success(self, mock_load_and_validate_message, mock_call_task, message_data, message):
        mock_load_and_validate_message.return_value = message
        receipt = str(uuid.uuid4())
        message_handler(json.dumps(message_data), receipt)
        mock_load_and_validate_message.assert_called_once_with(message_data)
        mock_call_task.assert_called_once_with(message)

    def test_fails_on_invalid_json(self, *mocks):
        with pytest.raises(ValueError):
            message_handler("bad json", None)

    def test_fails_on_validation_error(self, mock_load_and_validate_message, mock_call_task, message_data):
        error_message = 'Invalid message body'
        mock_load_and_validate_message.side_effect = ValidationError(error_message)
        with pytest.raises(ValidationError):
            message_handler(json.dumps(message_data), None)
        mock_call_task.assert_not_called()

    def test_fails_on_task_failure(self, mock_load_and_validate_message, mock_call_task, message_data, message):
        mock_load_and_validate_message.return_value = message
        mock_call_task.side_effect = Exception
        with pytest.raises(mock_call_task.side_effect):
            message_handler(json.dumps(message_data), None)

    def test_post_deserialize_hook(
            self, mock_load_and_validate_message, mock_call_task, message_data, message, settings):
        settings.HEDWIG_POST_DESERIALIZE_HOOK = 'tests.test_consumer.post_deserialize_hook'

        mock_load_and_validate_message.return_value = message
        receipt = str(uuid.uuid4())
        message_handler(json.dumps(message_data), receipt)
        mock_load_and_validate_message.assert_called_once_with(message_data)
        mock_call_task.assert_called_once_with(message)

        post_deserialize_hook.assert_called_once_with(message_data=message_data)

    def test_special_handling_logging_error(self, mock_load_and_validate_message, mock_call_task, message_data, message):
        mock_load_and_validate_message.return_value = message
        mock_call_task.side_effect = LoggingException('foo', extra={'mickey': 'mouse'})
        with pytest.raises(LoggingException), mock.patch.object(consumer.logger, 'exception') as logging_mock:
            message_handler(json.dumps(message_data), None)

            logging_mock.assert_called_once_with('foo', extra={'mickey': 'mouse'})

    def test_special_handling_retry_error(self, mock_load_and_validate_message, mock_call_task, message_data, message):
        mock_load_and_validate_message.return_value = message
        mock_call_task.side_effect = RetryException
        with pytest.raises(mock_call_task.side_effect), mock.patch.object(consumer.logger, 'info') as logging_mock:
            message_handler(json.dumps(message_data), None)

            logging_mock.assert_called_once()

    def test_special_handling_ignore_exception(self, mock_load_and_validate_message, mock_call_task, message_data,
                                               message):
        mock_load_and_validate_message.return_value = message
        mock_call_task.side_effect = IgnoreException
        # no exception raised
        with mock.patch.object(consumer.logger, 'info') as logging_mock:
            message_handler(json.dumps(message_data), None)

            logging_mock.assert_called_once()


@mock.patch('hedwig.consumer.message_handler', autospec=True)
def test_message_handler_sqs(mock_message_handler):
    queue_message = mock.MagicMock()
    message_handler_sqs(queue_message)

    mock_message_handler.assert_called_once_with(queue_message.body, queue_message.receipt_handle)


@mock.patch('hedwig.consumer.message_handler', autospec=True)
def test_message_handler_lambda(mock_message_handler):
    lambda_event = mock.MagicMock()
    message_handler_lambda(lambda_event)

    mock_message_handler.assert_called_once_with(lambda_event['Sns']['Message'], None)


def test_get_queue_messages():
    queue = mock.MagicMock()
    num_messages = 2
    visibility_timeout = 100

    messages = get_queue_messages(queue, num_messages, visibility_timeout=visibility_timeout)

    queue.receive_messages.assert_called_once_with(
        MaxNumberOfMessages=num_messages,
        WaitTimeSeconds=WAIT_TIME_SECONDS,
        MessageAttributeNames=['All'],
        VisibilityTimeout=visibility_timeout,
    )
    assert messages == queue.receive_messages.return_value


def test_get_default_queue_name():
    assert get_default_queue_name() == f'HEDWIG-{settings.HEDWIG_QUEUE.upper()}'


pre_process_hook = mock.MagicMock()
post_deserialize_hook = mock.MagicMock()


@mock.patch('hedwig.consumer.get_queue_messages', autospec=True)
@mock.patch('hedwig.consumer.message_handler_sqs', autospec=True)
class TestFetchAndProcessMessages:
    def test_success(self, mock_message_handler, mock_get_messages):
        queue_name = 'my-queue'
        queue = mock.MagicMock()
        num_messages = 3
        visibility_timeout = 4

        mock_get_messages.return_value = [mock.MagicMock(), mock.MagicMock()]

        fetch_and_process_messages(queue_name, queue, num_messages, visibility_timeout)

        mock_get_messages.assert_called_once_with(queue, num_messages, visibility_timeout=visibility_timeout)
        mock_message_handler.assert_has_calls([mock.call(x) for x in mock_get_messages.return_value])
        for message in mock_get_messages.return_value:
            message.delete.assert_called_once_with()

    def test_preserves_messages(self, mock_message_handler, mock_get_messages):
        queue_name = 'my-queue'
        queue = mock.MagicMock()

        mock_get_messages.return_value = [mock.MagicMock()]
        mock_message_handler.side_effect = Exception

        fetch_and_process_messages(queue_name, queue)

        mock_get_messages.return_value[0].delete.assert_not_called()

    def test_ignore_delete_error(self, mock_message_handler, mock_get_messages):
        queue_name = 'my-queue'
        queue = mock.MagicMock()

        mock_get_messages.return_value = [mock.MagicMock()]
        mock_get_messages.return_value[0].delete.side_effect = Exception

        with mock.patch.object(consumer.logger, 'exception') as logging_mock:
            fetch_and_process_messages(queue_name, queue)

            logging_mock.assert_called_once()

        mock_get_messages.return_value[0].delete.assert_called_once_with()

    def test_pre_process_hook(self, mock_message_handler, mock_get_messages, settings):
        queue_name = 'my-queue'
        queue = mock.MagicMock()
        settings.HEDWIG_PRE_PROCESS_HOOK = 'tests.test_consumer.pre_process_hook'

        mock_get_messages.return_value = [mock.MagicMock(), mock.MagicMock()]

        fetch_and_process_messages(queue_name, queue)

        pre_process_hook.assert_has_calls([
            mock.call(sqs_queue_message=x)
            for x in mock_get_messages.return_value
        ])


@mock.patch('hedwig.consumer.message_handler_lambda', autospec=True)
class TestProcessMessagesForLambdaConsumer:
    def assert_extra_logged(self, logging_mock, expected):
        actual = logging_mock.call_args[1]['extra']
        assert actual == expected

    def test_success(self, mock_message_handler):
        # copy from https://docs.aws.amazon.com/lambda/latest/dg/eventsources.html#eventsources-sns
        mock_record1 = {
            "EventVersion": "1.0",
            "EventSubscriptionArn": "arn",
            "EventSource": "aws:sns",
            "Sns": {
                "SignatureVersion": "1",
                "Timestamp": "1970-01-01T00:00:00.000Z",
                "Signature": "EXAMPLE",
                "SigningCertUrl": "EXAMPLE",
                "MessageId": "95df01b4-ee98-5cb9-9903-4c221d41eb5e",
                "Message": "Hello from SNS!",
                "MessageAttributes": {
                    "request_id": {
                        "Type": "String",
                        "Value": str(uuid.uuid4())
                    },
                    "TestBinary": {
                        "Type": "Binary",
                        "Value": "TestBinary"
                    }
                },
                "Type": "Notification",
                "UnsubscribeUrl": "EXAMPLE",
                "TopicArn": "arn",
                "Subject": "TestInvoke"
            }
        }
        mock_record2 = {
            "EventVersion": "1.0",
            "EventSubscriptionArn": "arn",
            "EventSource": "aws:sns",
            "Sns": {
                "SignatureVersion": "1",
                "Timestamp": "1970-01-01T00:00:00.000Z",
                "Signature": "EXAMPLE",
                "SigningCertUrl": "EXAMPLE",
                "MessageId": "95df01b4-ee98-5cb9-9903-4c221d41eb5e",
                "Message": "Hello from SNS!",
                "MessageAttributes": {
                    "request_id": {
                        "Type": "String",
                        "Value": str(uuid.uuid4())
                    },
                    "TestBinary": {
                        "Type": "Binary",
                        "Value": "TestBinary"
                    }
                },
                "Type": "Notification",
                "UnsubscribeUrl": "EXAMPLE",
                "TopicArn": "arn",
                "Subject": "TestInvoke"
            }
        }
        event = {
            "Records": [
                mock_record1,
                mock_record2
            ]
        }
        process_messages_for_lambda_consumer(event)
        mock_message_handler.assert_has_calls([
            mock.call(mock_record1),
            mock.call(mock_record2),
        ])


@mock.patch('hedwig.consumer.get_queue', autospec=True)
@mock.patch('hedwig.consumer.fetch_and_process_messages', autospec=True)
class TestListenForMessages:
    def test_listen_for_messages(self, mock_fetch_and_process, mock_get_queue):
        num_messages = 3
        visibility_timeout_s = 4
        loop_count = 1

        listen_for_messages(num_messages, visibility_timeout_s, loop_count)

        queue_name = get_default_queue_name()
        mock_get_queue.assert_called_once_with(queue_name)
        mock_fetch_and_process.assert_called_once_with(
            queue_name, mock_get_queue.return_value, num_messages=num_messages,
            visibility_timeout=visibility_timeout_s,
        )



from distutils.version import StrictVersion
import random
import time
from unittest import mock
import uuid

import pytest

from hedwig.conf import settings
from hedwig.exceptions import ValidationError, CallbackNotFound
from hedwig.models import Message, MessageType, Metadata, _get_sqs_client
from hedwig.testing.factories import MetadataFactory


class TestMetadata:
    def test_new(self):
        data = MetadataFactory()
        metadata = Metadata(data)
        assert metadata.timestamp == data['timestamp']
        assert metadata.headers == data['headers']
        assert metadata.publisher == data['publisher']
        assert metadata.receipt is None

    def test_equal(self):
        metadata = Metadata(MetadataFactory())
        assert metadata == metadata

    def test_equal_fail(self):
        metadata = Metadata(MetadataFactory())
        assert metadata != metadata.as_dict()


class TestMessageMethods:
    publisher = 'myapi'

    @mock.patch('hedwig.models.time.time', autospec=True)
    def test_create_metadata(self, mock_time):
        mock_time.return_value = time.time()
        headers = {
            'foo': 'bar',
        }

        assert Message._create_metadata(headers) == {
            'timestamp': int(mock_time.return_value * 1000),
            'publisher': settings.HEDWIG_PUBLISHER,
            'headers': headers,
        }

    def test_constructor(self):
        message_data = {
            "id": str(uuid.uuid4()),
            "metadata": {
                "timestamp": 1460868253255,
                "publisher": "myapp",
                "headers": {
                    "request_id": str(uuid.uuid4()),
                },
            },
            "format_version": "1.0",
            "schema": "https://hedwig.automatic.com/schema#/schemas/trip_created/1.0",
            "data": {
                "vehicle_id": 'C_1234567890123456',
                "user_id": 'U_1234567890123456',
            }
        }

        message = Message(message_data)
        assert message.id == message_data['id']
        assert message.metadata == Metadata(message_data['metadata'])
        assert message.headers == message_data['metadata']['headers']
        assert message.schema == 'https://hedwig.automatic.com/schema#/schemas/trip_created/1.0'
        assert message.data == message_data['data']

    def test_new(self, message_data):
        message = Message.new(
            MessageType.trip_created, StrictVersion('1.0'), message_data['data'], message_data['id'],
            message_data['metadata']['headers'],
        )

        assert message.id == message_data['id']
        assert message.format_version == Message.FORMAT_CURRENT_VERSION
        assert message.headers == message_data['metadata']['headers']
        assert message.schema == 'https://hedwig.automatic.com/schema#/schemas/trip_created/1.0'
        assert message.data == message_data['data']

    def test_as_dict(self, message):
        assert message.as_dict() == {
            'format_version': str(message.format_version),
            'id': message.id,
            'metadata': message.metadata.as_dict(),
            'schema': message.schema,
            'data': message.data,
        }

    def test_validate_callback(self, message):
        message.validate()
        message.validate_callback()
        assert message.callback is not None

    @pytest.mark.parametrize('missing_data', ['id', 'metadata', 'format_version', 'schema', 'data'])
    def test_validate_missing_data(self, missing_data, message_data):
        message_data[missing_data] = None

        with pytest.raises(ValidationError):
            Message(message_data)

    def test_validate_invalid_schema(self, message):
        message._schema = 'foo'
        with pytest.raises(ValidationError):
            message.validate()

    def test_validate_invalid_version(self, message_data):
        message_data['format_version'] = '3.0'

        with pytest.raises(ValidationError):
            Message(message_data)

    @mock.patch('hedwig.callback.Callback.find_by_message', side_effect=CallbackNotFound)
    def test_validate_missing_task(self, _, message):
        message.validate()
        with pytest.raises(ValidationError):
            message.validate_callback()

    @mock.patch('tests.handlers._trip_created_handler', autospec=True)
    def test_exec_callback(self, mock_trip_created_handler, message):
        message.validate()
        message.validate_callback()
        message.metadata.receipt = str(uuid.uuid4())
        message.exec_callback()
        mock_trip_created_handler.assert_called_once_with(message)

    @mock.patch('hedwig.models._get_sqs_client')
    @mock.patch('hedwig.consumer.get_default_queue_name')
    def test_extend_visibility_timeout(self, mock_get_queue_name, mock_get_sqs_client, message):
        visibility_timeout_s = random.randint(0, 1000)

        message.extend_visibility_timeout(visibility_timeout_s)

        mock_get_queue_name.assert_called_once_with()
        mock_get_sqs_client.assert_called_once_with()
        mock_get_sqs_client.return_value.get_queue_url.assert_called_once_with(
            QueueName=mock_get_queue_name.return_value
        )
        mock_get_sqs_client.return_value.change_message_visibility.assert_called_once_with(
            QueueUrl=mock_get_sqs_client.return_value.get_queue_url.return_value['QueueUrl'],
            ReceiptHandle=message.receipt,
            VisibilityTimeout=visibility_timeout_s,
        )

    def test_equal(self, message):
        assert message == message

    def test_equal_fail(self, message):
        assert message != message.as_dict()

    def test_getter_timestamp(self, message):
        assert message.timestamp == message.metadata.timestamp

    def test_getter_publisher(self, message):
        assert message.publisher == message.metadata.publisher


@mock.patch('hedwig.models.boto3.client')
def test_get_sqs_client(mock_boto3_client):
    client = _get_sqs_client()
    mock_boto3_client.assert_called_once_with(
        'sqs',
        region_name=settings.AWS_REGION,
        aws_access_key_id=settings.AWS_ACCESS_KEY,
        aws_secret_access_key=settings.AWS_SECRET_KEY,
        aws_session_token=settings.AWS_SESSION_TOKEN,
        endpoint_url=settings.AWS_ENDPOINT_SQS,
    )
    assert client == mock_boto3_client.return_value
